from .base import registered_device_types  # noqa
from .kettle_redmond import RedmondKettle  # noqa
from .xiaomi_ht import XiaomiHumidityTemperatureV1  # noqa
from .xiaomi_lywsd03 import XiaomiHumidityTemperatureLYWSD  # noqa
